{
  "cells": [
    {
      "cell_type": "markdown",
      "metadata": {},
      "source": [
        "\n# TSNE\u4e2d\u7684\u8fd1\u4f3c\u6700\u8fd1\u90bb\n\n\u672c\u793a\u4f8b\u5c55\u793a\u4e86\u5982\u4f55\u5728\u7ba1\u9053\u4e2d\u94fe\u63a5KNeighborsTransformer\u548cTSNE\u3002\n\u5b83\u8fd8\u5c55\u793a\u4e86\u5982\u4f55\u5305\u88c5 `nmslib` \u548c `pynndescent` \u5305\u4ee5\u66ff\u6362KNeighborsTransformer\u5e76\u6267\u884c\u8fd1\u4f3c\u6700\u8fd1\u90bb\u3002\u8fd9\u4e9b\u5305\u53ef\u4ee5\u901a\u8fc7 `pip install nmslib pynndescent` \u8fdb\u884c\u5b89\u88c5\u3002\n\n\u6ce8\u610f\uff1a\u5728KNeighborsTransformer\u4e2d\uff0c\u6211\u4eec\u4f7f\u7528\u7684\u5b9a\u4e49\u5305\u62ec\u6bcf\u4e2a\u8bad\u7ec3\u70b9\u4f5c\u4e3a\u5176\u81ea\u8eab\u7684\u90bb\u5c45\u5728 `n_neighbors` \u7684\u8ba1\u6570\u4e2d\uff0c\u5e76\u4e14\u4e3a\u4e86\u517c\u5bb9\u6027\u539f\u56e0\uff0c\u5f53 `mode == 'distance'` \u65f6\u4f1a\u8ba1\u7b97\u4e00\u4e2a\u989d\u5916\u7684\u90bb\u5c45\u3002\n\u8bf7\u6ce8\u610f\uff0c\u6211\u4eec\u5728\u63d0\u8bae\u7684 `nmslib` \u5305\u88c5\u4e2d\u4e5f\u505a\u4e86\u540c\u6837\u7684\u5904\u7406\u3002\n"
      ]
    },
    {
      "cell_type": "code",
      "execution_count": null,
      "metadata": {
        "collapsed": false
      },
      "outputs": [],
      "source": [
        "# \u4f5c\u8005\uff1ascikit-learn \u5f00\u53d1\u8005\n# SPDX-License-Identifier: BSD-3-Clause"
      ]
    },
    {
      "cell_type": "markdown",
      "metadata": {},
      "source": [
        "\u9996\u5148\u6211\u4eec\u5c1d\u8bd5\u5bfc\u5165\u8fd9\u4e9b\u5305\uff0c\u5e76\u5728\u7f3a\u5931\u65f6\u8b66\u544a\u7528\u6237\u3002\n\n"
      ]
    },
    {
      "cell_type": "code",
      "execution_count": null,
      "metadata": {
        "collapsed": false
      },
      "outputs": [],
      "source": [
        "import sys\n\ntry:\n    import nmslib\nexcept ImportError:\n    print(\"The package 'nmslib' is required to run this example.\")\n    sys.exit()\n\ntry:\n    from pynndescent import PyNNDescentTransformer\nexcept ImportError:\n    print(\"The package 'pynndescent' is required to run this example.\")\n    sys.exit()"
      ]
    },
    {
      "cell_type": "markdown",
      "metadata": {},
      "source": [
        "\u6211\u4eec\u5b9a\u4e49\u4e86\u4e00\u4e2a\u5305\u88c5\u7c7b\uff0c\u7528\u4e8e\u5b9e\u73b0 `nmslib` \u7684 scikit-learn API\uff0c\u4ee5\u53ca\u4e00\u4e2a\u52a0\u8f7d\u51fd\u6570\u3002\n\n"
      ]
    },
    {
      "cell_type": "code",
      "execution_count": null,
      "metadata": {
        "collapsed": false
      },
      "outputs": [],
      "source": [
        "import joblib\nimport numpy as np\nfrom scipy.sparse import csr_matrix\n\nfrom sklearn.base import BaseEstimator, TransformerMixin\nfrom sklearn.datasets import fetch_openml\nfrom sklearn.utils import shuffle\n\n\nclass NMSlibTransformer(TransformerMixin, BaseEstimator):\n    \"\"\"\u7528\u4e8e\u5c06nmslib\u4f5c\u4e3asklearn\u7684KNeighborsTransformer\u7684\u5305\u88c5\u5668\"\"\"\n\n    def __init__(self, n_neighbors=5, metric=\"euclidean\", method=\"sw-graph\", n_jobs=-1):\n        self.n_neighbors = n_neighbors\n        self.method = method\n        self.metric = metric\n        self.n_jobs = n_jobs\n\n    def fit(self, X):\n        self.n_samples_fit_ = X.shape[0]\n\n        # \u8bf7\u53c2\u9605\u624b\u518c\u4e2d\u7684\u66f4\u591a\u6307\u6807\n        # https://github.com/nmslib/nmslib/tree/master/manual\n        space = {\n            \"euclidean\": \"l2\",\n            \"cosine\": \"cosinesimil\",\n            \"l1\": \"l1\",\n            \"l2\": \"l2\",\n        }[self.metric]\n\n        self.nmslib_ = nmslib.init(method=self.method, space=space)\n        self.nmslib_.addDataPointBatch(X.copy())\n        self.nmslib_.createIndex()\n        return self\n\n    def transform(self, X):\n        n_samples_transform = X.shape[0]\n\n        # \u51fa\u4e8e\u517c\u5bb9\u6027\u539f\u56e0\uff0c\u7531\u4e8e\u6bcf\u4e2a\u6837\u672c\u90fd\u88ab\u89c6\u4e3a\u5176\u81ea\u8eab\u7684\u90bb\u5c45\uff0c\u56e0\u6b64\u5c06\u591a\u8ba1\u7b97\u4e00\u4e2a\u989d\u5916\u7684\u90bb\u5c45\u3002\n        n_neighbors = self.n_neighbors + 1\n\n        if self.n_jobs < 0:\n            # \u4e0e joblib \u5bf9\u8d1f n_jobs \u503c\u7684\u5904\u7406\u76f8\u540c\uff1a\n            # \u7279\u522b\u5730\uff0c `n_jobs == -1` \u8868\u793a\u201c\u7ebf\u7a0b\u6570\u4e0e CPU \u6570\u76f8\u540c\u201d\u3002\n            num_threads = joblib.cpu_count() + self.n_jobs + 1\n        else:\n            num_threads = self.n_jobs\n\n        results = self.nmslib_.knnQueryBatch(\n            X.copy(), k=n_neighbors, num_threads=num_threads\n        )\n        indices, distances = zip(*results)\n        indices, distances = np.vstack(indices), np.vstack(distances)\n\n        indptr = np.arange(0, n_samples_transform * n_neighbors + 1, n_neighbors)\n        kneighbors_graph = csr_matrix(\n            (distances.ravel(), indices.ravel(), indptr),\n            shape=(n_samples_transform, self.n_samples_fit_),\n        )\n\n        return kneighbors_graph\n\n\ndef load_mnist(n_samples):\n    \"\"\"\u52a0\u8f7dMNIST\uff0c\u6253\u4e71\u6570\u636e\uff0c\u53ea\u8fd4\u56den_samples\u3002\"\"\"\n    mnist = fetch_openml(\"mnist_784\", as_frame=False)\n    X, y = shuffle(mnist.data, mnist.target, random_state=2)\n    return X[:n_samples] / 255, y[:n_samples]"
      ]
    },
    {
      "cell_type": "markdown",
      "metadata": {},
      "source": [
        "\u6211\u4eec\u5bf9\u4e0d\u540c\u7684\u7cbe\u786e/\u8fd1\u4f3c\u6700\u8fd1\u90bb\u53d8\u6362\u5668\u8fdb\u884c\u57fa\u51c6\u6d4b\u8bd5\u3002\n\n"
      ]
    },
    {
      "cell_type": "code",
      "execution_count": null,
      "metadata": {
        "collapsed": false
      },
      "outputs": [],
      "source": [
        "import time\n\nfrom sklearn.manifold import TSNE\nfrom sklearn.neighbors import KNeighborsTransformer\nfrom sklearn.pipeline import make_pipeline\n\ndatasets = [\n    (\"MNIST_10000\", load_mnist(n_samples=10_000)),\n    (\"MNIST_20000\", load_mnist(n_samples=20_000)),\n]\n\nn_iter = 500\nperplexity = 30\nmetric = \"euclidean\"\n# TSNE \u9700\u8981\u4e00\u5b9a\u6570\u91cf\u7684\u90bb\u5c45\uff0c\u8fd9\u53d6\u51b3\u4e8e\u56f0\u60d1\u5ea6\u53c2\u6570\u3002\n# \u7531\u4e8e\u6211\u4eec\u5c06\u6bcf\u4e2a\u6837\u672c\u90fd\u5305\u542b\u4e3a\u5176\u81ea\u8eab\u7684\u90bb\u5c45\uff0c\u56e0\u6b64\u9700\u8981\u589e\u52a0\u4e00\u4e2a\u3002\nn_neighbors = int(3.0 * perplexity + 1) + 1\n\ntsne_params = dict(\n    init=\"random\",  # pca not supported for sparse matrices\n    perplexity=perplexity,\n    method=\"barnes_hut\",\n    random_state=42,\n    n_iter=n_iter,\n    learning_rate=\"auto\",\n)\n\ntransformers = [\n    (\n        \"KNeighborsTransformer\",\n        KNeighborsTransformer(n_neighbors=n_neighbors, mode=\"distance\", metric=metric),\n    ),\n    (\n        \"NMSlibTransformer\",\n        NMSlibTransformer(n_neighbors=n_neighbors, metric=metric),\n    ),\n    (\n        \"PyNNDescentTransformer\",\n        PyNNDescentTransformer(\n            n_neighbors=n_neighbors, metric=metric, parallel_batch_queries=True\n        ),\n    ),\n]\n\nfor dataset_name, (X, y) in datasets:\n    msg = f\"Benchmarking on {dataset_name}:\"\n    print(f\"\\n{msg}\\n\" + str(\"-\" * len(msg)))\n\n    for transformer_name, transformer in transformers:\n        longest = np.max([len(name) for name, model in transformers])\n        start = time.time()\n        transformer.fit(X)\n        fit_duration = time.time() - start\n        print(f\"{transformer_name:<{longest}} {fit_duration:.3f} sec (fit)\")\n        start = time.time()\n        Xt = transformer.transform(X)\n        transform_duration = time.time() - start\n        print(f\"{transformer_name:<{longest}} {transform_duration:.3f} sec (transform)\")\n        if transformer_name == \"PyNNDescentTransformer\":\n            start = time.time()\n            Xt = transformer.transform(X)\n            transform_duration = time.time() - start\n            print(\n                f\"{transformer_name:<{longest}} {transform_duration:.3f} sec\"\n                \" (transform)\"\n            )"
      ]
    },
    {
      "cell_type": "markdown",
      "metadata": {},
      "source": [
        "\u793a\u4f8b\u8f93\u51fa\uff1a\n\n## \u5728MNIST_10000\u4e0a\u7684\u57fa\u51c6\u6d4b\u8bd5\uff1a\nKNeighborsTransformer  0.007 \u79d2\uff08\u62df\u5408\uff09\nKNeighborsTransformer  1.139 \u79d2\uff08\u8f6c\u6362\uff09\nNMSlibTransformer      0.208 \u79d2\uff08\u62df\u5408\uff09\nNMSlibTransformer      0.315 \u79d2\uff08\u8f6c\u6362\uff09\nPyNNDescentTransformer 4.823 \u79d2\uff08\u62df\u5408\uff09\nPyNNDescentTransformer 4.884 \u79d2\uff08\u8f6c\u6362\uff09\nPyNNDescentTransformer 0.744 \u79d2\uff08\u8f6c\u6362\uff09\n\n## \u5728MNIST_20000\u4e0a\u7684\u57fa\u51c6\u6d4b\u8bd5\uff1a\nKNeighborsTransformer  0.011 \u79d2\uff08\u62df\u5408\uff09\nKNeighborsTransformer  5.769 \u79d2\uff08\u8f6c\u6362\uff09\nNMSlibTransformer      0.733 \u79d2\uff08\u62df\u5408\uff09\nNMSlibTransformer      1.077 \u79d2\uff08\u8f6c\u6362\uff09\nPyNNDescentTransformer 14.448 \u79d2\uff08\u62df\u5408\uff09\nPyNNDescentTransformer 7.103 \u79d2\uff08\u8f6c\u6362\uff09\nPyNNDescentTransformer 1.759 \u79d2\uff08\u8f6c\u6362\uff09\n\n\u8bf7\u6ce8\u610f\uff0c `PyNNDescentTransformer` \u5728\u7b2c\u4e00\u6b21 `fit` \u548c\u7b2c\u4e00\u6b21 `transform` \u65f6\u4f1a\u82b1\u8d39\u66f4\u591a\u65f6\u95f4\uff0c\u8fd9\u662f\u7531\u4e8e numba \u5373\u65f6\u7f16\u8bd1\u5668\u7684\u5f00\u9500\u3002\u4f46\u5728\u7b2c\u4e00\u6b21\u8c03\u7528\u4e4b\u540e\uff0c\u7f16\u8bd1\u540e\u7684 Python \u4ee3\u7801\u4f1a\u88ab numba \u7f13\u5b58\u5728\u7f13\u5b58\u4e2d\uff0c\u540e\u7eed\u8c03\u7528\u4e0d\u4f1a\u518d\u6709\u8fd9\u79cd\u521d\u59cb\u5f00\u9500\u3002:class:`~sklearn.neighbors.KNeighborsTransformer` \u548c `NMSlibTransformer` \u5728\u8fd9\u91cc\u53ea\u8fd0\u884c\u4e00\u6b21\uff0c\u56e0\u4e3a\u5b83\u4eec\u7684 `fit` \u548c `transform` \u65f6\u95f4\u66f4\u7a33\u5b9a\uff08\u5b83\u4eec\u6ca1\u6709 PyNNDescentTransformer \u7684\u51b7\u542f\u52a8\u95ee\u9898\uff09\u3002\n\n"
      ]
    },
    {
      "cell_type": "code",
      "execution_count": null,
      "metadata": {
        "collapsed": false
      },
      "outputs": [],
      "source": [
        "import matplotlib.pyplot as plt\nfrom matplotlib.ticker import NullFormatter\n\ntransformers = [\n    (\"TSNE with internal NearestNeighbors\", TSNE(metric=metric, **tsne_params)),\n    (\n        \"TSNE with KNeighborsTransformer\",\n        make_pipeline(\n            KNeighborsTransformer(\n                n_neighbors=n_neighbors, mode=\"distance\", metric=metric\n            ),\n            TSNE(metric=\"precomputed\", **tsne_params),\n        ),\n    ),\n    (\n        \"TSNE with NMSlibTransformer\",\n        make_pipeline(\n            NMSlibTransformer(n_neighbors=n_neighbors, metric=metric),\n            TSNE(metric=\"precomputed\", **tsne_params),\n        ),\n    ),\n]\n\n# \u521d\u59cb\u5316\u56fe\u8868\nnrows = len(datasets)\nncols = np.sum([1 for name, model in transformers if \"TSNE\" in name])\nfig, axes = plt.subplots(\n    nrows=nrows, ncols=ncols, squeeze=False, figsize=(5 * ncols, 4 * nrows)\n)\naxes = axes.ravel()\ni_ax = 0\n\nfor dataset_name, (X, y) in datasets:\n    msg = f\"Benchmarking on {dataset_name}:\"\n    print(f\"\\n{msg}\\n\" + str(\"-\" * len(msg)))\n\n    for transformer_name, transformer in transformers:\n        longest = np.max([len(name) for name, model in transformers])\n        start = time.time()\n        Xt = transformer.fit_transform(X)\n        transform_duration = time.time() - start\n        print(\n            f\"{transformer_name:<{longest}} {transform_duration:.3f} sec\"\n            \" (fit_transform)\"\n        )\n\n        # \u7ed8\u5236 TSNE \u5d4c\u5165\u56fe\uff0c\u8be5\u56fe\u5728\u4e0d\u540c\u65b9\u6cd5\u4e4b\u95f4\u5e94\u8be5\u975e\u5e38\u76f8\u4f3c\n        axes[i_ax].set_title(transformer_name + \"\\non \" + dataset_name)\n        axes[i_ax].scatter(\n            Xt[:, 0],\n            Xt[:, 1],\n            c=y.astype(np.int32),\n            alpha=0.2,\n            cmap=plt.cm.viridis,\n        )\n        axes[i_ax].xaxis.set_major_formatter(NullFormatter())\n        axes[i_ax].yaxis.set_major_formatter(NullFormatter())\n        axes[i_ax].axis(\"tight\")\n        i_ax += 1\n\nfig.tight_layout()\nplt.show()"
      ]
    },
    {
      "cell_type": "markdown",
      "metadata": {},
      "source": [
        "\u793a\u4f8b\u8f93\u51fa\uff1a\n\n## \u5728MNIST_10000\u4e0a\u7684\u57fa\u51c6\u6d4b\u8bd5\uff1a\n\u4f7f\u7528\u5185\u90e8NearestNeighbors\u7684TSNE 24.828\u79d2\uff08fit_transform\uff09\n\u4f7f\u7528KNeighborsTransformer\u7684TSNE 20.111\u79d2\uff08fit_transform\uff09\n\u4f7f\u7528NMSlibTransformer\u7684TSNE 21.757\u79d2\uff08fit_transform\uff09\n\n## \u5728MNIST_20000\u4e0a\u7684\u57fa\u51c6\u6d4b\u8bd5\uff1a\n\u4f7f\u7528\u5185\u90e8NearestNeighbors\u7684TSNE 51.955\u79d2\uff08fit_transform\uff09\n\u4f7f\u7528KNeighborsTransformer\u7684TSNE 50.994\u79d2\uff08fit_transform\uff09\n\u4f7f\u7528NMSlibTransformer\u7684TSNE 43.536\u79d2\uff08fit_transform\uff09\n\n\u6211\u4eec\u53ef\u4ee5\u89c2\u5bdf\u5230\uff0c\u9ed8\u8ba4\u7684 :class:`~sklearn.manifold.TSNE` \u4f30\u8ba1\u5668\u53ca\u5176\u5185\u90e8\u7684 :class:`~sklearn.neighbors.NearestNeighbors` \u5b9e\u73b0\uff0c\u5728\u6027\u80fd\u65b9\u9762\u5927\u81f4\u7b49\u540c\u4e8e\u5305\u542b :class:`~sklearn.manifold.TSNE` \u548c :class:`~sklearn.neighbors.KNeighborsTransformer` \u7684\u7ba1\u9053\u3002\u8fd9\u662f\u9884\u671f\u7684\uff0c\u56e0\u4e3a\u8fd9\u4e24\u4e2a\u7ba1\u9053\u5728\u5185\u90e8\u90fd\u4f9d\u8d56\u4e8e\u76f8\u540c\u7684 :class:`~sklearn.neighbors.NearestNeighbors` \u5b9e\u73b0\u6765\u6267\u884c\u7cbe\u786e\u7684\u90bb\u5c45\u641c\u7d22\u3002\u8fd1\u4f3c\u7684 `NMSlibTransformer` \u5728\u6700\u5c0f\u7684\u6570\u636e\u96c6\u4e0a\u5df2\u7ecf\u6bd4\u7cbe\u786e\u641c\u7d22\u7a0d\u5feb\uff0c\u4f46\u8fd9\u79cd\u901f\u5ea6\u5dee\u5f02\u9884\u8ba1\u5728\u6837\u672c\u6570\u91cf\u8f83\u5927\u7684\u6570\u636e\u96c6\u4e0a\u4f1a\u66f4\u52a0\u663e\u8457\u3002\n\n\u8bf7\u6ce8\u610f\uff0c\u5e76\u975e\u6240\u6709\u7684\u8fd1\u4f3c\u641c\u7d22\u65b9\u6cd5\u90fd\u80fd\u4fdd\u8bc1\u6bd4\u9ed8\u8ba4\u7684\u7cbe\u786e\u641c\u7d22\u65b9\u6cd5\u63d0\u9ad8\u901f\u5ea6\uff1a\u5b9e\u9645\u4e0a\uff0c\u81ea\u4ecescikit-learn 1.1\u4ee5\u6765\uff0c\u7cbe\u786e\u641c\u7d22\u7684\u5b9e\u73b0\u5df2\u7ecf\u663e\u8457\u6539\u8fdb\u3002\u6b64\u5916\uff0c\u86ee\u529b\u7cbe\u786e\u641c\u7d22\u65b9\u6cd5\u5728 `fit` \u65f6\u4e0d\u9700\u8981\u6784\u5efa\u7d22\u5f15\u3002\u56e0\u6b64\uff0c\u8981\u5728:class:`~sklearn.manifold.TSNE` \u7ba1\u9053\u7684\u4e0a\u4e0b\u6587\u4e2d\u83b7\u5f97\u6574\u4f53\u6027\u80fd\u63d0\u5347\uff0c\u8fd1\u4f3c\u641c\u7d22\u5728 `transform` \u65f6\u7684\u6536\u76ca\u9700\u8981\u5927\u4e8e\u5728 `fit` \u65f6\u6784\u5efa\u8fd1\u4f3c\u641c\u7d22\u7d22\u5f15\u6240\u82b1\u8d39\u7684\u989d\u5916\u65f6\u95f4\u3002\n\n\u6700\u7ec8\uff0cTSNE\u7b97\u6cd5\u672c\u8eab\u4e5f\u662f\u8ba1\u7b97\u5bc6\u96c6\u578b\u7684\uff0c\u4e0e\u6700\u8fd1\u90bb\u641c\u7d22\u65e0\u5173\u3002\u56e0\u6b64\uff0c\u5c06\u6700\u8fd1\u90bb\u641c\u7d22\u6b65\u9aa4\u52a0\u901f5\u500d\uff0c\u5e76\u4e0d\u4f1a\u4f7f\u6574\u4e2a\u6d41\u7a0b\u52a0\u901f5\u500d\u3002\n\n"
      ]
    }
  ],
  "metadata": {
    "kernelspec": {
      "display_name": "Python 3",
      "language": "python",
      "name": "python3"
    },
    "language_info": {
      "codemirror_mode": {
        "name": "ipython",
        "version": 3
      },
      "file_extension": ".py",
      "mimetype": "text/x-python",
      "name": "python",
      "nbconvert_exporter": "python",
      "pygments_lexer": "ipython3",
      "version": "3.11.9"
    }
  },
  "nbformat": 4,
  "nbformat_minor": 0
}